from __future__ import annotations

from audidata.datasets.audiocaps import AudioCaps
from audidata.datasets.clotho import Clotho
from audidata.datasets.clotho_aqa import ClothoAQA
from audidata.datasets.free_spoken_digit import FreeSpokenDigit
from audidata.datasets.gtzan import GTZAN
from audidata.datasets.librispeech import LibriSpeech
from audidata.datasets.ljspeech import LJSpeech
from audidata.datasets.maestro import MAESTRO
from audidata.datasets.musdb18hq import MUSDB18HQ
from audidata.datasets.shutterstock import Shutterstock
# from audidata.datasets.esc50 import ESC50
# from audidata.datasets.slakh2100 import Slakh2100
# from audidata.datasets.urmp import URMP
# from audidata.datasets.gmd import GMD
# from audidata.datasets.isophonics import Isophonics
from audidata.datasets.voicebank_demand import VoicebankDemand
# from audidata.datasets.magnatagatune import MagnaTagATune
from audidata.datasets.wavcaps import WavCaps
